module Make (Env : sig

  (* [lookup x] returns the pseudo-register that holds the local
     variable [x]. *)

  val lookup: string -> Register.t

  (* [allocate()] returns a fresh pseudo-register. *)

  val allocate: unit -> Register.t

  (* [generate instruction] returns a fresh instruction label, which
     it associates with [instruction] in the control flow graph. *)

  val generate: RTL.instruction -> Label.t

  (* [loop target] creates a fresh instruction label [label], which it
     associates with an unconditional branch instruction whose target
     is [target label]. It returns [target label]. *)

  val loop: (Label.t -> Label.t) -> Label.t

  (* [is_exit label] tells whether the label [label] is the exit label
     of the current procedure or function. This can be used to determine
     which calls are tail calls. *)

  val is_exit: Label.t -> bool

  (* [result] is [None] if this is a procedure, and [Some f] if this
     a function named [f]. *)

  val result: string option

end) = struct

open Integer
open MIPSOps
open Env

(* ------------------------------------------------------------------------- *)

(* Translating expressions. *)

(* [pick_destination e] decides which pseudo-register will hold the
   result of evaluating expression [e]. *)

let pick_destination e =
  match e with
  | _ ->
      allocate()

(* [translate_expression destr e destl] generates new instructions
   whose effect is to evaluate the expression [e], to place its value
   in the pseudo-register [destr], and to transfer control to the
   label [destl]. It returns the entry label of the newly generated
   instructions. *)

let rec translate_expression
   (destr : Register.t)
   (e : UPP.expression)
   (destl : Label.t)
    : Label.t =

  match e with

  | _ ->
      failwith "UNIMPLEMENTED. FIX ME!"

(* Translating function and procedure calls. This is analogous to the
   case of binary operator applications above, except the number of
   arguments is variable. The destination register is optional --
   present for functions, absent for procedures. *)

and translate_call 
   (odestr : Register.t option)
   (callee : Primitive.callee)
   (actuals : UPP.expression list)
   (destl : Label.t)
   : Label.t =

  let temporaries = List.map pick_destination actuals in
  List.fold_right2 translate_expression temporaries actuals (
  generate (RTL.ICall (odestr, callee, temporaries, destl))
  )

(* Translating tail calls. This is analogous to the case of ordinary calls
   above, except a tail call does not return, so there is no destination
   register and no successor label. *)

and translate_tail_call 
   (callee : Primitive.callee)
   (actuals : UPP.expression list)
   : Label.t =

  let temporaries = List.map pick_destination actuals in
  List.fold_right2 translate_expression temporaries actuals (
  generate (RTL.ITailCall (callee, temporaries))
  )

(* ------------------------------------------------------------------------- *)

(* [mkunbranch e uncon truel falsel] translates the expression [e],
   writing its value to a temporary register; then, it issues a unary
   branch instruction on that register, whose condition is [uncon],
   and whose target labels are [truel] and [falsel]. *)

let mkunbranch e uncon truel falsel =
  let temporary = pick_destination e in
  translate_expression temporary e (
  generate (RTL.IUnBranch (uncon, temporary, truel, falsel))
  )

(* [mkbinbranch e1 e2 bincon truel falsel] translates the expressions
   [e1] and [e2], writing their values to temporary registers; then,
   it issues a binary branch instruction on those registers, whose
   condition is [bincon], and whose target labels are [truel] and
   [falsel]. *)

let mkbinbranch e1 e2 bincon truel falsel =
  let temporary1 = pick_destination e1
  and temporary2 = pick_destination e2 in
  translate_expression temporary1 e1 (
  translate_expression temporary2 e2 (
  generate (RTL.IBinBranch (bincon, temporary1, temporary2, truel, falsel))
  ))

(* Translating conditions. *)

(* [translate_condition c truel falsel] generates new [RTL]
   instructions whose effect is to evaluate the condition [c] and to
   transfer control to one of the labels [truel] and [falsel],
   depending on the condition's value. It returns the entry label of
   the newly generated instructions. *)

let rec translate_condition
   (c : UPP.condition)
   (truel : Label.t)
   (falsel : Label.t)
    : Label.t =

  match c with

  | _ ->
      failwith "UNIMPLEMENTED. FIX ME!"

(* ------------------------------------------------------------------------- *)

(* Translating instructions. *)

(* [translate_instruction i destl] generates new [RTL] instructions
   whose effect is to execute the [UPP] instruction [i] and to
   transfer control to the destination label [destl]. It returns the
   entry label of the newly generated instructions. *)

let rec translate_instruction
   (i : UPP.instruction)
   (destl : Label.t)
    : Label.t =

  match i with

    (* Procedure calls. *)

  | UPP.IProcCall (callee, actuals) ->
      translate_call None callee actuals destl

    (* Local variable update. *)

  | UPP.ISetVar (x, e) ->
      failwith "UNIMPLEMENTED. FIX ME!"

    (* Global variable update. *)

  | UPP.ISetGlobal (offset, e) ->
      let temporary = pick_destination e in
      translate_expression temporary e (
        generate (RTL.ISetGlobal (offset, temporary, destl))
      )

    (* Memory write. This is translated to the corresponding [RTL]
       instruction. *)

  | UPP.IStore (eaddress, offset, evalue) ->
      let address = pick_destination eaddress
      and value = pick_destination evalue in
      translate_expression address eaddress (
      translate_expression value evalue (
      generate (RTL.IStore (address, offset, value, destl))
      ))

    (* Sequence. This is translated by chaining the destination
       labels. *)

  | UPP.ISeq instructions ->
      List.fold_right translate_instruction instructions destl

    (* Conditional. *)

  | UPP.IIf (c, i1, i2) ->
      failwith "UNIMPLEMENTED. FIX ME!"

    (* Loop. *)

  | UPP.IWhile (c, i) ->
      failwith "UNIMPLEMENTED. FIX ME!"

(* ------------------------------------------------------------------------- *)

end

